import numpy as np
import itertools
from diversity import memoized
from typing import List, Optional
from tqdm import tqdm
from rouge_score import rouge_scorer
from evaluate import load


def homogenization_score(
        data: List[str],
        measure: str = 'rougel',
        use_stemmer: Optional[str] = False,
        model: Optional[str] = 'distilbert-base-uncased',
        verbose: Optional[bool] = True

) -> float:
    """ Calculates the homogenization score for a set of documents (corpus-level). 
        From https://arxiv.org/pdf/2309.05196.pdf 
     Args:
         data (List[str]): Strings to score.
         measure (str, optional): Either 'rougel', 'bertscore', or 'bleu'. Defaults to 'rougel'.
         use_stemmer(str, optional): Whether to use stemming in the ROUGE-L calculation. Defaults to False.
         model(str, optional): Model to use for BERTScore. Defaults to 'microsoft/deberta-xlarge-mnli'. 
         verbose(bool, optional): Whether to display progress bar. Defaults to True.
     Returns:
         float: Homogenization score.
     """

    if measure == 'rougel':
        scorer = rouge_scorer.RougeScorer(['rougeL'], use_stemmer=use_stemmer)
    elif measure == 'bertscore': 
        scorer = load("bertscore")
    elif measure == 'bleu':
        scorer = load("bleu")
    else: 
        raise ValueError("Scoring measure must be one of `rougel`, `bleu`, or `bertscore`.")

    # all pairs of comparisons (cartesian product) without repetition at the same index
    all_pairs = ((data[i], data[j]) 
                    for i in range(len(data))
                    for j in range(len(data))
                    if i != j)

    curr_str = data[0]
    corpus_score = 0
    doc_score = 0
    
    if verbose:
        print('==> Scoring all pairs')
        for pair in tqdm(all_pairs, total=len(data)**2):
            # single document-level homogenization score, pairs are ordered
            if pair[0] == curr_str:     
                doc_score += _calculate_score(pair, scorer, measure, model)
            else:
                corpus_score += doc_score / (len(data) - 1)
                curr_str = pair[0]
                doc_score = 0 
                doc_score += _calculate_score(pair, scorer, measure, model)
    else:
        for pair in all_pairs:
            # single document-level homogenization score, pairs are ordered
            if pair[0] == curr_str:     
                doc_score += _calculate_score(pair, scorer, measure, model)
            else:
                corpus_score += doc_score / (len(data) - 1)
                curr_str = pair[0]
                doc_score = 0 
                doc_score += _calculate_score(pair, scorer, measure, model)
        
    # returns corpus level homogenization score 
    return round(corpus_score / len(data), 3)


@memoized
def _calculate_score(pair, scorer, measure, model):
    """ Returns the score of two strings """
    
    # repeated string (excat str match) that is *not* at the same position 
    if pair[0] == pair[1]: 
        return 1
    else:
        if measure == 'rougel':
            score = scorer.score(pair[0], 
                                 pair[1])['rougeL'][-1]
        elif measure == 'bertscore': 
            score = scorer.compute(predictions=[pair[0]], 
                                   references=[pair[1]], 
                                   model_type=model)['f1'][0]
        elif measure == 'bleu': 
            score = scorer.compute(predictions=[pair[0]], 
                                   references=[pair[1]])['bleu']
        
        return score 




from typing import List, Tuple
import nltk 


def token_patterns(
        data: List[str],
        n: int,
        top_n: int = 10
) -> List[Tuple[str, int]]:
    """ Finds ngrams patterns in the data.

    Args:
        data (List[str]): Data to run frequency counts on.
        n (int): N-gram length.
        top_n (int, optional): Top patterns to display. Defaults to 10.

    Returns:
        List[Tuple[str, int]]: Sorted list of top n-gram patterns.
    """

    # treat data as one string 
    all_data = ' '.join(data)

    ngrams = list(nltk.ngrams(all_data.split(' '), n))
    frequency = nltk.FreqDist(ngrams)

    sorted_frequency = sorted(frequency.items(), key=lambda kv: kv[1], reverse=True)[:top_n]

    sorted_frequency = [(' '.join(x[0]), x[1]) for x in sorted_frequency]

    return sorted_frequency



import collections
import functools

class memoized(object):
   '''
   Decorator. Caches a function's return value each time it is called.
   If called later with the same arguments, the cached value is returned
   (not re-evaluated).
   From https://wiki.python.org/moin/PythonDecoratorLibrary#Memoize.
   '''

   def __init__(self, func):
      self.func = func
      self.cache = {}

   def __call__(self, *args):
      if not isinstance(args, collections.abc.Hashable):
         # better to not cache than blow up.
         return self.func(*args)
      if args in self.cache:
         return self.cache[args]
      else:
         value = self.func(*args)
         self.cache[args] = value
         return value
      
   def __repr__(self):
      '''Return the function's docstring.'''
      return self.func.__doc__
   
   def __get__(self, obj, objtype):
      '''Support instance methods.'''
      return functools.partial(self.__call__, obj)



from .compression import compression_ratio
from .patterns.token import token_patterns
from .patterns.part_of_speech import pos_patterns, get_pos
from .utils.memoize import memoized
from .homogenization import homogenization_score
from .ngram_diversity import ngram_diversity_score
from .eigen_auc import eigen_auc



from typing import List, Tuple, Any, Set

import nltk
nltk.download('averaged_perceptron_tagger')

def get_pos(
        data: List[str]
) -> Tuple[List[str], List[Tuple[str, str]]]:
    """ Turns a sequence into parts of speech.

    Args:
        data (List[str]): Data to tranform into part of speech tags.

    Returns:
        Tuple[List[str], List[Tuple[str, str]]]: Part-of-speech tags only, tuple of (token, part-of-speech tag).
    """

    pos_tuples = [nltk.pos_tag(x.split()) for x in data]

    joined_pos = []
    joined_text = []

    for doc in pos_tuples:
        joined_text.append(' '.join([x[0] for x in doc]))
        joined_pos.append(' '.join([x[1] for x in doc]))

    return joined_pos, pos_tuples


def _find_sub_list(
        sl: List[Any],
        l: List[Any]
) -> List[Any]:
    """ Given a pattern and a list of strings, returns sublists matching the pattern. """
    
    results = []
    sll = len(sl)
    for ind in (i for i,e in enumerate(l) if e==sl[0]):
        if l[ind:ind+sll]==sl:
            results.append((ind, ind+sll-1))

    return results


def pos_patterns(
        text: List[List[Tuple[str, str]]],
        pattern: str
) -> Set[str]: 
    """ Finds all substrings matching a part of speech pattern. 

    Args:
        text (List[List[Tuple[str, str]]]): Text containing words and part-of-speech tags.
        pattern (str): Part-of-speech tag pattern to search for.

    Returns:
        Set[str]: Returns all the string matching the pattern.
    """

    pos = []
    word = []
    
    # text is a list of lists of tuples (word, part of speech)
    for doc in text: 
        pos.append([i[1] for i in doc])
        word.append([i[0] for i in doc])
    
    pos = [' '.join(x) for x in pos]
    word = [' '.join(x) for x in word]

    all_matches = [] 

    # return positions of each tag and the corresponding tokens
    for w, p in zip(word, pos):

        test = _find_sub_list(pattern.split(), p.split())

        if test:
            for occ in test: 
                splits = w.split()[int(occ[0]):int(occ[1]+1)]
                all_matches.append(" ".join(splits))

    return set(all_matches) 



import numpy as np
from typing import List

def eigen_auc(
        embeddings: List[np.ndarray]
) -> float:
    """Calculates the eigenAUC score for a set of token embeddings (corpus-level). 

    Args:
        embeddings (List[np.ndarray]): List of embeddings representing a single token for each datapoint.

    Returns:
        float: eigenAUC score.
    """
    
    data = np.vstack(embeddings)
    _, S, _ = np.linalg.svd(data, full_matrices=False)

    eigenvalues = S**2 
    variances = _variance_at_all_k(eigenvalues)

    auc = np.trapz(variances)

    return round(auc/len(variances), 3)


def _variance_explained(eigenvalues, k):
    """ Retrieves the variance explained by top-k eigenvalues. """
    total_variance = np.sum(eigenvalues)
    top_k_variance = np.sum(eigenvalues[:k])

    return top_k_variance / total_variance


def _variance_at_all_k(eigenvalues):
    """ Gets variances for all values of k. """
    variances = []

    for k in range(1, len(eigenvalues)):
        var = _variance_explained(eigenvalues, k)
        variances.append(var)

    return variances



from typing import List, Optional
from pathlib import Path

import tempfile
import gzip
import os
import lzma as xz


def compression_ratio(
        data: List[str],
        algorithm: str = 'gzip',
        verbose: bool = False,
        path: Optional[str] = None
) -> float:
    """ Calculates the compression ratio for a collection of text.
     Args:
         path (str): Path to store temporarily zipped files.
         data (List[str]): Strings to compress.
         algorithm (str, optional): Either 'gzip' or 'xz'. Defaults to 'gzip'.
         verbose (bool, optional): Print out the original and compressed size separately. Defaults to False.
     Returns:
         float: Compression ratio (original size / compressed size)
     """
     
    temp_dir = None
    if not path:
        temp_dir = tempfile.TemporaryDirectory()
        path = Path(temp_dir.name)
    else:
        path = Path(path)

    with (path / 'original.txt').open('w+') as f:
        f.write(' '.join(data))

    original_size = os.path.getsize(os.path.join(path, "original.txt"))

    if algorithm == 'gzip':

        with gzip.GzipFile(str(path / 'compressed.gz'), 'w+') as f:
            f.write(gzip.compress(' '.join(data).encode('utf-8')))

        compressed_size = os.path.getsize(os.path.join(path, "compressed.gz"))

    elif algorithm == 'xz': 

        with xz.open(str(path / 'compressed.gz'), 'wb') as f:
            f.write(' '.join(data).encode('utf-8'))

        compressed_size = (path / "compressed.gz").stat().st_size

    if verbose: 
        print(f"Original Size: {original_size}\nCompressed Size: {compressed_size}")

    if temp_dir:
        temp_dir.cleanup()

    return round(original_size / compressed_size, 3)
    



from typing import List
import nltk

def ngram_diversity_score(
        data: List[str],
        num_n: int = 4, 
) -> float:
    """ Calculates corpus-level ngram diversity based on unique ngrams 
       (e.g., https://arxiv.org/pdf/2202.00666.pdf).

    Args:
        data (List[str]): List of documents. 
        num_n (int): Max ngrams to test up to. Defaults to 5. 

    Returns:
        float: ngram diveristy score.
    """
    score = 0 
    data = ' '.join(data).split(' ') # format to list of words

    for i in range(1, num_n + 1): 
        ngrams = list(nltk.ngrams(data, i))
        # num unique ngrams / all ngrams for each size n 
        score += len(set(ngrams)) / len(ngrams) 

    return round(score, 3)
